#!/usr/sbin/perl -w

# Florence Corpet INRA 7avril98 - 31mars00
#DisplayFam is a browser for the analysis of multiple alignments of large numbers
#of protein sequences.
#
#DisplayFam users must cite:
#F. Corpet, J. Gouzy and D. Kahn (1999), 'Browsing protein families via the "Rich 
#Family Description" data structure.', Bioinformatics, 15:12, 1020-1027.
#
# Permission granted to use and modify this script so long as the
# copyright above is maintained, modifications are documented, and
# credit is given for any use of the script.
#
#For more information e-mail to proquest@toulouse.inra.fr

#
###################################################
# this file includes three parts
#	1: a personalisable part
#	2: a common part (flo_proc.pl)
#	3: a common part
#	4: a default main program as an example

# installation
# edit part 1 with your own values
# copy cgi-lib2.pl in this file directory.
# copy all the following files in a same directory ($PROG)
# 	standalone programs called : protdist, bionj, TriangleTree, reroot, consens, subtree, drawtree
# 	data files : species, speciesCG
# copy the following images in a same directory (which URL is $HTTPIMAGE)
# 	images (.gif): plus, redball, triangle, (yellow, blue, red, green, black)square, alignment, domains, tree
###################################################

###################################################
#### 1: a personalisable part
#### Look at all comment lines and edit as you wish
###################################################

#### Choose whether you include stylesheet ($CSS=1) or not ($CSS=0) in the Html pages
$CSS=0;

#### InitFlo : init global variables for URLs and disk path
#### you must personalise these strings for a cgi-bin utilisation
#### $PROG: program and working directory
#### $DOMDIR: data directory
#### $HTTPDATA: data directory url
#### $HTTPIMAGE: standard image directory url
#### $SUBTREEPL: URL of the main script that calls DisplayFam
####
#### you must personalise this program call
#### $FETCHFAM: program that prints the family ID_FAM to stdout
####(at run-time ID_FAM is substituted by $id_fam)
#### if $FETCHFAM uses data files, be careful to use absolute file names or
#### names relative to the working directory ($PROG)
#### you can use the msf2rfd function to translate msf data to
#### "Rich Family Description" data
#### $rfd_multi-line_string = &msf2rfd ($msf_multi-line_string);
####
#### you can personalise these strings
#### $Sorry: warning message when id_fam is not valid

sub InitFlo
{
	my ($HTTPROOT, $HTTPSCRIPT, $DATA);
	my $loc = !$ENV{'SERVER_NAME'};

# Init PROG, DOMDIR, HTTPDATA, HTTPIMAGE

	#$HTTPROOT="../..";
	#$HTTPSCRIPT="";
	#$HTTPIMAGE="../images/";
	
	#if ($loc) 	############  file:// request ###############
	#{
	#	$DOMDIR="./";
	#	$PROG="../DisplayFam/";
	#	$_ = `pwd`;
	#	chomp;
	#	$HTTPDATA="file:$_/";
	#	$HTTPIMAGE=$HTTPDATA;
	#} 
	#else 		############ http:// request  ###############	 
	#{

		#<----------- Some lines in our apache configuration file (srm.conf) related to DisplayFam configuration --------->

		# DocumentRoot /www
		# ScriptAlias /DisplayFam/cgi-bin/ /www/prodom/DisplayFam/
		# AddHandler imap-file cf

		#<----------- DisplayFam web configuration starts here ---------------->

		# tree 
		# /www/					WWW Server root
		# /www/prodom/DisplayFam/		DisplayFam main directory
		# /www/prodom/DisplayFam/tmp/		DisplayFam temporary directory, the www daemon needs RW access on this directory
		# /www/prodom/DisplayFam/images/	images directory

		# /www/		Server root
		$ROOT="/www/";

		######	Access from the web ######

		$PROG="/prodom/DisplayFam/";

		# /www/prodom/DisplayFam/tmp/           temporary directory, the www daemon needs RW access
		$HTTPDATA= $PROG."tmp/";
		#   /prodom/DisplayFam/images/	images directory
		$HTTPIMAGE=$PROG."images/";

		######	Access from system ######

		# /www/prodom/DisplayFam/tmp/		temporary directory, the www daemon needs RW access
		$DOMDIR   =$ROOT.$PROG."/tmp/";

		# /www/prodom/DisplayFam/
		$PROG=$ROOT.$PROG;

		#<----------- DisplayFam web configuration ends here ---------------->

	#}

# Init programs
	$FETCHFAM="cat ${DOMDIR}ID_FAM.in";
	$SUBTREEPL="DisplayFam.pl";

# Your initialisations
	$Project="";
	if ($id_fam eq "new")
	{
		$id_fam=time;
		while (&ExistFile($DOMDIR.$id_fam) or &ExistFile("${DOMDIR}${id_fam}.in"))
		{
			sleep(1);
			$id_fam= time;
		}
		$Ac_fam=$id_fam;
		if ($Alignment ne "")
		{
			open (INSEQ, ">${DOMDIR}${id_fam}.in");
			print INSEQ &msf2rfd($Alignment);
			close (INSEQ);
			$Alignment="";
		}
	}
	$Sorry= "Sorry, your request is not valid";

# First time, init the frames
	if ($wanted eq 'align')
	{ 
		print STDOUT "<frameset rows='30%,70%'>\n";
		print STDOUT "	<frame name='${Project}Domain_$Ac_fam' src='".&Script(0,"descript")."'>\n";
		print STDOUT "	<frame name='${Project}Alignment_$Ac_fam' src='".&Script(0,"alignment")."'>\n";
		print STDOUT "</frameset></html>\n";
		exit();
	}
	print STDOUT "<BODY $BODY>\n";
}

#### getRef($AC) returns the url of the link that uses $AC
#### the return string will replace GETREF in <A HREF = "GETREF">$AC_sequence_label</A>
#### e.g. return "http://www.expasy.ch/cgi-bin/get-sprot-entry?$AC"
sub getRef
{
	my ($AC) = @_;
	if ( length($AC) > 10 )
	{
		my(@F)=split('_',$AC,2);
		return "http://www.expasy.ch/cgi-bin/get-sprot-entry?$F[0]";
	}
	else
	{
		return "http://www.expasy.ch/cgi-bin/get-sprot-entry?$AC";
	}
}


##########################################################################
## printBigTitle : a Title for all windows
##
## @Title @Bottom can be used
## @Title includes the lines that were before the sequences in the original file
##	@Bottom includes the lines that were after the sequences in the original file
## e.g. print STDOUT "<H2>Example Family</H2><BR>";
### default : empty function
##########################################################################
sub printBigTitle
{
	local $_ = join("\n",@Title);
	my $t = &getDescription;
	$t = "Example Family" unless $t;
	print STDOUT "<H2>$t</H2><BR>";
	my $first = 1;
	print STDOUT $TableBegin;
	foreach $_ (@Title)
	{
		if (&isIDline) {}
		elsif (&isSequenceNumberLine)
		{
			print STDOUT &TableRow("Number of sequences in family",&getSequenceNumber), "\n";
		}
		elsif (&isAlignmentLengthLine)
		{
			print STDOUT &TableRow("Alignment length",&getAlignmentLength), "\n";
		}
		elsif (&isDescriptionLine)
		{
			if ($first)
			{
				$first=0;# the first description line is the BigTitle line.
				next;
			}
			$t = &getDescription;
			if ($t=~/:/)
			{
				print STDOUT &TableRow($`,$'), "\n";
			}
			else
			{
				print STDOUT &TableRow("Comment",$t), "\n";
			}
		}
	}
	print STDOUT $TableEnd, "\n";
}

######### setSubtitles sets 2 titles for the alignment frame and the tree window
## setSubtitles($name) : if ($name) it is a sub-family
## $Subtitle is a true title 
## $Subtitle2 is initialised to SubTitle2Mask that is the following string
## "Other descriptions of the (sub)family corresponding to this alignment: tree-reference, PERSO"
## in this function:
## you must set $Subtitle 
## you must replace PERSO in $Subtitle2 by a list of references to other descriptions of the family
## you can use the Ahref and Image functions
### 
sub setSubtitles
{
	my ($name) = @_;
	$Subtitle = '';
	$Subtitle2 =~ s/PERSO//;
	if ($name)
	{
		$Subtitle= &CenterText("Subfamily: $name and ".($NbCons-1)." other similar sequences");
	}
}

######### printSites prints a special line with the sequence format
### to underline some positions in the alignment, for example.
### default : empty function
sub printSites
{
}

######## printBottom prints lines after the alignment
### default : empty function
sub printBottom
{
}

### printSpecialForm prints extra information to the parameter form
### e.g. print STDOUT "<input name='my_name' type='hidden' value='my_value'>";
sub printSpecialForm
{
}

### addSpecialParam returns extra parameters to the main script command line
### must return a string like "name=value&amp;" or ""
### e.g. return "my_name='my_value'&amp;";
sub addSpecialParam
{
	return "";
}

# Set the $release variable to your value. It wil be used for temporary file names
# and can be used in the following functions
sub InitRelease
{
	$release = "";
}

# ValidDom check if $id is valid
sub ValidDom
{
	my($id)=@_;
  return	( $id ne "" );
}

# getAC_ID returns AC and ID from the sequence label
# AC is used to link the sequence label with a sequence server
# ID is used as the sequence label to display
# for drawtree to handle the IDs correctly they MUST include one and only one
# underscore char (_)
# e.g. sequence label format can be AC|ID with ID= PROT_SPECIES as in SwissProt
# return split('|', shift)
# or the ID can be used as an AC: $AC = $ID = $label; return ($AC,$ID);
sub getAC_ID
{
	my ($label) = @_;
	my ($AC,$ID);
#	return split('\|',$label);# | between AC and ID
	$AC = $ID = $label; return ($AC,$ID);# no AC in the label
}

# correctAC_ID ($AC, $ID) builds a correct label with an AC and
# an $ID that does not include exactly one underscore
sub correctAC_ID
{
	my ($AC, $ID) = @_;
	my @G = split ('_',$ID,2);
	return $ID if (@G == 2);
	if (scalar(@G)<2)
	{
		$ID = "_$ID";# the Class info is after the _
#		$ID .="_";# the Class info is before the _
	}
	else
	{
		$G[1]=~ s/_.*//;# delete info from the second _ to the end
#		$G[1]=~ s/_/-/g ;# keeps the info but use - instead of _
		$ID="$G[0]_$G[1]";
	}
#	return "$AC|$ID";# | between AC and ID
	return $ID;# no AC in the label
}

# getIDSp gets the part of the ID that is used to define its Class
sub getIDSp
{
	my ($ID) = @_;
	my (@G) = split('_',$ID);
#	return $G[0]; # Class info is the first word
	return $G[0] unless ($G[1]);# in case of no Class info, returns the other word
	return $G[1];# Class info is the second word
}

# InitExtraInfo init the name of the extra info file $SPECIES
sub InitExtraInfo
{
	$SPECIES = "species";
}

##################### end of the personalisable part

##############################
#### 2: a common part (flo_proc.pl)
##############################
###########################################################
#
# general tools
#
###########################################################

sub ExistFile
{
  my($name)=@_;
  my($ret_val)=(0);
    if ( (-e $name) && !(-z $name) )
    {
        $ret_val = 1;
    }
    return ($ret_val);
}

sub max
{
	my($n1,$n2) = @_;
	($n1 > $n2) ? $n1 : $n2;
}

###########################################################
#
# functions related to our rfd format
#
###########################################################

# block delimiter
$DomainEnd = "//\n";
sub isEnd
{
	m!$DomainEnd!;
}
sub putEnd
{
	$DomainEnd;
}

# ID domain_ID
sub isIDline
{
	/^ID/i;
}
sub getID
{
 my ($id) = /^ID\s+(\S+)/mi;
 $id;
}
sub putID
{
		"ID $_[0]\n";
}

# AC PDxxxxxx
sub isACline
{
	/^AC/i;
}
sub getAC
{
	my ($ac) = /^AC\s+(PD0*\d+)/mi;
	$ac;
}
sub putAC
{
	"AC $_[0]\n";
}

# DE description (optional)
sub isDescriptionLine
{
	/^DE/i;
}
sub getDescription
{
 my ($description)=/^DE\s+(.*)/mi;
 $description;
}
sub putDE
{
	"DE $_[0]";
}

# 'CC -!- key: value' is a parsable comment line
# list of keys
$TREE_METHOD_KEY = "TREE BUILDING";
sub isParsableComment 
{
	/^CC\s+\-\!\-\s+([^:]+)\:\s+(.*)/i;
}

sub isParsableCommentWithKey
{
	my ($key) = @_;
	return 0 unless &isParsableComment;
	my ($new_key, $new_val) = &getParsableComment;
	print STDERR "$new_key=? $key\n";
	return ($new_key eq $key);
}

sub getParsableComment
{
	my($key,$val)=/^CC\s+\-\!\-\s+([^:]+)\:\s+(.*)/;
	return ($key,$val);
}
sub putParsableComment
{
	"CC   -!- $_[0]: $_[1]\n";
}

# DC database comment
sub isDatabaseComment
{
	/^DC/i;
}
sub getDatabaseComment
{
	my ($dc) = /^DC\s+(.*)/mi; 
	return $dc;
}

# KW list of keywords
sub isKeywordLine
{
	/^KW/i;
}
sub getKW
{
 my ($keywords)=/^KW\s+(.*)/mi;
 split(/[; ]+/, $keywords);
}
sub putKW
{
	"KW ". join(' ',@_)."\n";
}

# DR database links
sub isDataRelationLine
{
	/^DR/i;
}
$getDR = "^DR\\s+([^ ;]+);\\s+(.*)";
sub getDR
{
 my ($database, $links) = /$getDR/mi;
 my @links = split(/[; ]+/, $links);
 ($database, @links);
}
sub putDR
{
	my ($database, @links) = @_;
	"DR   $database;         ". join(' ',@links)."\n";
}

# TR tree
sub isTreeLine
{
	/^TR/i;
}
sub getTR
{
	join ("\n",/^TR\s+(.*)/mig);
}

sub putTR
{
	"TR ".join ("\nTR ",@_)."\n";
}

# AL sequence_name begin end weight sequence
# CP sequence_number sequence_name weight sequence (partial consensus)
sub isSequenceLine
{
	/^(AL|CP)/i;
}
$getALCP="^(AL|CP\\s+\\d+)\\s+(\\S+)(\\s+\\d+\\s+\\d+\\s+|\\s+)(\\d*\\.*\\d+)\\s+(\\S.*)";
sub getAL
{
	my ($seqName, $beg, $end, $w, $ali) =  /^AL\s+(\S+)\s+(\d+)\s+(\d+)\s+([\d\.]+)\s+(\S.*)/mi;
	$ali =~ tr/ //d;
	$ali =~ tr/a-z/A-Z/;
	($seqName, $beg, $end, $w, $ali);
}
sub getCP
{
	my ($nb, $seqName, $w, $ali) = /^CP\s+(\d+)\s+(\S+)\s+([\d\.]+)\s+(\S.*)/mi;
	$ali =~ tr/ //d;
   	$ali =~ tr/a-z/A-Z/;
	($nb, $seqName, $w, $ali);
}
sub getALCP
{
	my ($tnb,$name,$begend,$w,$ali)=@_;
#  print STDERR join(",",@_),"\n";
	my ($nb, $beg, $end);
	if ($tnb =~ /^AL/i)
	{
		$nb =1;
		($beg,$end) = $begend=~ /\s+(\d+)\s+(\d+)/;
	}
	else
	{
		($nb) = $tnb=~ /(\d+)/;
		$beg = $end = '.';
	}
	$ali =~ tr/ //d;
   	$ali =~ tr/a-z/A-Z/;
	($nb,$name,$beg,$end,$w,$ali);
}
sub putAL
{
	my ($Seq) = @_;
	"AL $$Seq{name} $$Seq{begin} $$Seq{end} $$Seq{weight} $$Seq{alignment}\n"
}
sub putCP
{
	my ($Seq) = @_;
	"CP $$Seq{nb} $$Seq{name} $$Seq{weight} $$Seq{alignment}\n"
}

# CO consensus sequence
sub isConsensusLine
{
	/^CO/i;
}
sub getCO
{
 my ($consensus) = /^CO\s+(\S.*)/mi;
 $consensus;
}
sub putCO
{
	"CO $_[0]\n";
}

# LA alignment length
sub isAlignmentLengthLine
{
	/^LA/i;
}
sub getAlignmentLength
{
 my ($la)=/^LA\s+(\d+)/mi;
 $la;
}
sub putLA
{
	"LA $_[0]\n";
}

# LD mean number of amino acids in the domain (optional)
sub isLDline
{
	/^LD/i;
}
sub getLD
{
 my ($ld)=/^LD\s+(\d+)/mi;
 $ld;
}
sub putLD
{
	"LD $_[0]\n";
}


# ND number of aligned domains
sub isSequenceNumberLine
{
	/^ND/i;
}
sub getSequenceNumber
{
 my ($nd)=/^ND\s+(\d+)/mi;
 $nd;
}
sub putND
{
	"ND $_[0]\n";
}


# NP number of proteins in the family (< ND because of repetitions)
sub getNP
{
 my ($np)=/^NP\s+(\d+)/mi;
 $np;
}
sub putNP
{
	"NP $_[0]\n";
}


# lines that are not AL, CP, TR nor DR  are comment
# they can be split into two parts by the position of the firs AL line

sub getCommentOnly
{
	local ($_) = @_;
	local ($`, $');
	s/^AL.*\n//mgi;
	s/^CP.*\n//mgi;
	s/^TR.*\n//mgi;
	s/^DR.*\n//mgi;
	s!$DomainEnd!!mgi;
	split("\n");
}
sub getComment
{
	my ($t1, $t2)=@_;
	/^AL.*\n/mi;
	@$t1 = getCommentOnly($`);
	@$t2 = getCommentOnly($');
}

# getSeqID returns the sequence label to display from the sequence label as it
# is in the file
sub getSeqID
{
	my ($AC,$ID)= &getAC_ID(@_);
	$ID;
}

# getSeqAC returns an AC to query a sequence server from the sequence label
sub getSeqAC
{
	my ($AC,$ID)= &getAC_ID(@_);
	$AC;
}

# getSeqSp returns the Class info from the sequence label
sub getSeqSp
{
	return &getIDSp(&getSeqID(@_));
}

sub getMSFname
{
	my ($name, $len, $weight)=@_;
   return "" if ($name =~ /consens/i);
	push @Names, $name;
	return "AL $name 1 $len $weight ";
}

sub msf2rfd
{
	local ($_) = @_;
	my ($i, $t, $s, $n, $ss, @Lines);
	local @Names;
	if (/\.\.\s*$/mi)
	{
		$t = $`;
		$_ = $';
		$t =~ s/^/DE /mg;
	}
	push @Lines, &getMSFname($1,$2,$3)
		while (/Name:\s+(\S+)\s+Len:\s+(\d+).*Weight:\s+([\d\.]+)/mig);
	foreach $i (@Lines)
	{
  		$i=~tr/a-z/A-Z/;
		next if $i=~ s/^AL (\S+)\D(\d+)\-(\d+) 1 \d+/AL $1 $2 $3/;
		$i=~ s/^AL (\S+)\((\d+)\-(\d+)\) 1 \d+/AL $1 $2 $3/;
	}
	return unless (/^\/\//mi);
	$_=$';
	$i=0;
	while (/(\S+)\s+(.*)/mig)
	{
		$n=$1; $ss=$2;
		$s = $Names[$i];
		$s =~ s/(\W)/\\$1/g;
		next unless ($n =~ /$s/i);
		if ($n gt $Names[$i])
		{
			$Lines[$i]=~ s/^AL \S+ \d+ (\d+)/AL $n 1 $1/;
			$Lines[$i]=~ tr/a-z/A-Z/;
			$Lines[$i]=~ s/^AL (\S+)\((\d+)\-(\d+)\) 1 \d+/AL $1 $2 $3/
				unless $Lines[$i]=~ s/^AL (\S+)\D(\d+)\-(\d+) 1 \d+/AL $1 $2 $3/;
			$Names[$i]=$n;
		}
		$ss =~ tr/ //d;
		$Lines[$i] .= $ss;
		$i++;
		$i=0 if ($i > $#Names);
	}
	return join("\n",$t,@Lines)."\n//\n";
}

###################################
# domain data internal organisation
###################################
# comment lines are stored in @Title and @Bottom
# one is written before the alignment, the other one after.

# domain format
# $Id_fam family identificator
# $Length is the alignment length
# $Consensus is the original consensus sequence (not used ?)
# $NbSeq is the number of aligned domains
# @Sequences is an array of
# {
#	nb => $nb,
#	name => $seqName,
#	begin => $beg,
#	end => $end,
#	weight => $w,
#	alignment => $ali
# }

# tree format : tree is not used directly in perl => keeps as a string
# $Arbre = "(((1:0.64264,(((2:0.61888,3:0.67096):0.06406,..."

# @DataRelated is an array of links to databases

sub getSubfamily
{
	local ($_) = @_;
	(/(\d+)/g);
}

sub AddSequence
{
	my ($nb,$seqName,$beg, $end,$w,$ali)=@_;
   $NbSeq++ if ($nb < 2);
	push @Sequences, {"nb"=>$nb, "name"=>$seqName, "begin"=>$beg, "end"=>$end,
		"weight"=>$w, "alignment"=>$ali};
#	print STDERR scalar(@Sequences)," ";
}

sub AddDR
{
	my ($name,$data)=@_;
	$DataRelated{$name}=[] unless $DataRelated{$name};
	$name = $DataRelated{$name};
	push @$name, [split (' ',$data)];
}

sub Read1Domain
{
#	my ($name);
# init all global domain variables
	splice @Title, 0;
	splice @Bottom, 0;
	splice @Sequences, 0;
	%DataRelated = ();
	$Arbre = "";
	$NbSeq = 0;
	$Id_fam = &getID;
	$Ac_fam = &getAC;
	$Length = &getAlignmentLength;
	$Arbre = &getTR;
#	&AddSequence(&getALCP($1,$2,$3,$4,$5)) while (/$getALCP/mig);
	@_ = (/$getALCP/mig);
	while (@_)
	{
		&AddSequence(&getALCP(shift,shift,shift,shift,shift));
	}
	$Consensus = &getCO;
#	&AddDR ($1,$2) while (/$getDR/mig);
	@_ = (/$getDR/mig);
	while (@_)
	{
		&AddDR(shift,shift);
	}
	&getComment(*Title,*Bottom);
}

sub ReadDomain
{
# init all global domain variables
	my($file) = @_;
	local $/ = $DomainEnd;
	open (DOMAIN,$file) or return 0;
	$_ = <DOMAIN>;
	return 0 unless $_;
	&Read1Domain;
	close(DOMAIN);
	return !(($file =~ /\|/) && $?);
}

sub VerifyDomain
{
	local $/ = "\n";
	my ($NDline, @dummy) = grep &isSequenceNumberLine, @Title;
	return 0 if (@dummy);
	if ($NDline)
	{
		$_ = $NDline;
		# FS return 0 if ($NbSeq != &getSequenceNumber);
		if ( $NbSeq != ($nb=&getSequenceNumber) )
		{
			print STDOUT "ND=$nb \t nb_seq_in_alignment=$NbSeq\n";
			return (0);
		}
	}
	else
	{
		$_ = &putND($NbSeq);
		chomp;
		push @Title, $_;
	}

	my $Seq = $Sequences[0]{alignment};
	$Seq =~ tr/()[]//d;
	$Seq = length($Seq);
	if ($Length)
	{
		return 0 if ($Length != $Seq);
	}
	else
	{
		$Length = $Seq;
		$_ = &putLA($Length);
		chomp;
		push @Title, $_;
	}
	$Ac_fam = $Id_fam unless $Ac_fam;
	1;
}

sub Print1Domain
{
	my ($db, $links, $dr, $Seq);
	local $_ ;
	$_ = join ("\n",@Title)."\n" if (@Title);
	$_ .= &putTR(split("\n",$Arbre)) if ($Arbre);
	foreach $Seq (@Sequences)
	{
		if ($Seq->{nb} > 1)
		{
			$_.= &putCP($Seq);
		}
		else
		{
			$_ .= &putAL($Seq);
		}
	}
	$_ .= join ("\n",@Bottom)."\n" if (@Bottom);
	while (($db, $links) = each (%DataRelated))
	{
		foreach $dr (@$links)
		{
			$_ .= &putDR($db, @$dr);
		}
	}
	$_ .= &putEnd;
}

sub PrintDomain
{
	my ($file) = @_;
	open (DOMAIN,$file);
	print DOMAIN &Print1Domain;
	close DOMAIN;
}

sub CleanConsensus
{
	my $i;
	for ($i=0;$i < @Sequences;$i++)
	{
		next if ($Sequences[$i]{nb} < 2);
		splice (@Sequences, $i, 1);
		$i--;
	}
	$NbSeq = @Sequences;
}

sub ReadOrder_Weight
{
	my $i=0;
	my @OldSeq = @Sequences;
	$_= join("",@_);
	s/\n//g;
	@_ = split;
	while (@_)
	{
		$Sequences[$i] = $OldSeq[shift];
		$Sequences[$i]{weight} = shift;
		$i++;
	}
}

sub CompareStr
{
	my ($Seq1, $Seq2) = @_;
	my($Res,$length, $ident, $i);
	$Res = $Seq1 ^ $Seq2;# pos with identical chars become \0
	$ident = $Res =~ tr/\0//;#count identical chars
	$Seq1 =~ tr/A-Z/A/;
	$Seq2 =~ tr/A-Z/A/;
	$Res = $Seq1 ^ $Seq2;# pos with 2 letters become \0
	$length = $Res =~ tr/\0//;#count pos with 2 letters
	($ident,$length) = (-1,1) if ($length < 10);
	return ($ident,$length);
}

@dayhoff_pams = (
  195,   # 75.0% observed d; 195 PAMs estimated = 195% estimated d */
  196,   # 75.1% observed d; 196 PAMs estimated */
						197,    198,    199,    200,    200,    201,    202,  203,
  204,    205,    206,    207,    208,    209,    209,    210,    211,  212,
  213,    214,    215,    216,    217,    218,    219,    220,    221,  222,
  223,    224,    226,    227,    228,    229,    230,    231,    232,  233,
  234,    236,    237,    238,    239,    240,    241,    243,    244,  245,
  246,    248,    249,    250,    # 250 PAMs = 80.3% observed d */
											 252,    253,    254,    255,    257,  258,
  260,    261,    262,    264,    265,    267,    268,    270,    271,  273,
  274,    276,    277,    279,    281,    282,    284,    285,    287,  289,
  291,    292,    294,    296,    298,    299,    301,    303,    305,  307,
  309,    311,    313,    315,    317,    319,    321,    323,    325,  328,
  330,    332,    335,    337,    339,    342,    344,    347,    349,  352,
  354,    357,    360,    362,    365,    368,    371,    374,    377,  380,
  383,    386,    389,    393,    396,    399,    403,    407,    410,  414,
  418,    422,    426,    430,    434,    438,    442,    447,    451,  456,
  461,    466,    471,    476,    482,    487,    493,    498,    504,  511,
  517,    524,    531,    538,    545,    553,    560,    569,    577,  586,
  595,    605,    615,    626,    637,    649,    661,    675,    688,  703,
  719,    736,    754,    775,    796,    819,    845,    874,    907,  945,
			# 92.9% observed; 945 PAMs */
  988    # 93.0% observed; 988 PAMs */
);

sub PAMdist
{
	my ($ident) = @_;
	return -1 if ($ident < 0);
	return 0 if $ident >= 1;
	my $tt = 1.0 - $ident;
	if ($tt < 0.75)
	{
		$tt = -log(1-$tt-0.2*$tt*$tt);
	}
	else
	{
		if ($tt > 0.93)
		{
			$tt = 10.0;
		}
		else
		{
			$tt = 0.01*$dayhoff_pams[int (1000*$tt - 750)];
		}
	}
	return $tt;
}

sub Protdist
{
	my ($ret, $ident, $length, $Seq, @dist, $j);
	foreach $i (0..$#Sequences)
	{
		$dist[$i][$i] = '0.00';
		$Seq = $Sequences[$i]{alignment};
		$Seq =~ tr/A-Z/g/c;#code gaps as 'g';
		foreach $j ($i+1 .. $#Sequences)
		{
			($ident, $length) = &CompareStr($Seq,$Sequences[$j]{alignment});
			$ident = 0.01 * int(100*&PAMdist($ident/$length)+0.5);
			$ret = 1 if (($dist[$j][$i]=$dist[$i][$j]= $ident)<0);
		}
	}
	return ($ret, @dist);
}

sub Make2Tree
{
#	local $/ = "\n";
#	my $cde = "printf \"2\\n1 ".shift(@_).
#		"\\n2 ".shift(@_)."\\n\" |./protdist |";
	my ($d, @dist);
#	open (TREE,$cde) or return;
#	$_=<TREE>;
#	$_=<TREE>;
#	($d) = /\S+\s+[\.\d]+\s+(\d*\.\d{0,5})/;
	($d, @dist) = &Protdist;
	$d = $dist[0][1];
	if ($d == 0)
	{
		$d = "0.00000";
	}
	else
	{
		$d = ($d + 0.000018)/2;
		chop($d);
	}
	$Arbre = "(1:$d,2:$d);";
#	close(TREE) or ($Arbre="");
	$Arbre ="(1:10.0000,2:10.0000);" if ($?!=0);  # infinite distance
}

sub MakeTree
{
	my ($tmp) = ("$DOMDIR/maketree$release.$Id_fam");
	my $Seq;
	local $/ = "";
	&CleanConsensus if ($NbSeq  < @Sequences);
	$Arbre = "";
	$TreeMethod = -1;
	if (@Sequences < 3)
	{
		foreach $Seq (@Sequences)
		{
			$Seq->{weight}=1.00;
		}
		&Make2Tree ($Sequences[0]{alignment},$Sequences[1]{alignment})
			if (@Sequences == 2);
	}
	else
	{
		if (&ExistFile($tmp))
		{
		$TreeMethod = `grep -c "\-" $tmp`;
		$TreeMethod = ($TreeMethod > 0)?2:1;
#		print STDERR "use $tmp (method=$TreeMethod)\n";
		}
		else
		{
		open (SEQDATA,"| ./protdist >$tmp") or return;
#		open (SEQDATA,">$tmp") or return;
		print SEQDATA "$NbSeq\n";
		for $i (0..$#Sequences)
		{
			print SEQDATA "$i ",$Sequences[$i]{alignment},"\n";
		}
		close(SEQDATA);
		$TreeMethod = $?>>8;
		}
		if ($TreeMethod > 1)
		{
			print STDERR "TriangleTree used for $Id_fam\n";
			unless (open (TREE, "./TriangleTree $tmp | ./reroot |"))
			{
				$TreeMethod=3;
				print STDERR "can't compute a complete matrix distance\n";
				return(-1);
			}

		}
		else
		{
			open(TREE,"./bionj $tmp| ./reroot |")
				or return;
		}
		$Arbre = <TREE>;
		&ReadOrder_Weight(<TREE>);
		close(TREE);
		print STDERR "pb in maketree $Id_fam\n" if ($Arbre eq "");
# if some distances were estimated or if NbSeq is very large, keep the distance array		
		system ("rm -f $tmp") unless (($TreeMethod==1)or($NbSeq > 500));
	}
}

sub EditGap
{
	my ($seq) = @_;
	my ($j, $k);
	return $seq unless ($seq=~ /[\.\-]/);
	($j,$seq,$k) = $seq=~ /(^[\.\-]*)([A-Z\(\[].*[A-Z\)\]])([\.\-]*$)/;
	$j=~ tr/-/./;
	$seq=~ tr/./-/;
	$k=~ tr/-/./;
	join("",$j,$seq,$k);
}

sub Consens
{
	local $/ = "\n";
	my $tmp = "$DOMDIR/consens$release.$Id_fam";
	my $SW = $Sequences[0]{name} =~ /\|/;# to avoid XXXXX species
	my @OldSeq;
	my ($n, $i, $j, $k, $w, $seq, $name, $AC, $ID);

	&CleanConsensus if ($NbSeq  < @Sequences);
	return if ($NbSeq < 2);
#	print STDERR "ici : $tmp\n";
	open (TREE,"| ./consens >$tmp") or return;
#	open (TREE,">$tmp");
	print TREE "$NbSeq\n";
	$Arbre = "(1:10.0000,2:10.0000);\n" if ($Arbre eq "");
	print TREE "$Arbre\n";
	for $i (0..$#Sequences)
	{
		$Sequences[$i]{alignment} = &EditGap($Sequences[$i]{alignment});
		print TREE "$i ",$Sequences[$i]{weight}," ",$Sequences[$i]{alignment},"\n";
#		print STDERR "$i ",$Sequences[$i]{weight}," ",$Sequences[$i]{alignment},"\n";
	}
	close(TREE);
	return if ($?);
	open (CONS,"$tmp") or return;
	@OldSeq = splice (@Sequences, 0);
	while (defined($_= <CONS>))
	{
		chop;
		($n, $i, $w, $seq)=split(' ',$_,4);
		if ($n>1)
		{
			$name=$OldSeq[$i]{name};# cluster name = first seq name
			if ($SW)# search a better name for the cluster
			{
				for ($j=0, $k=-1;$j<$n;$j++)
				{
					($AC,$ID) = split ('\|',$OldSeq[$i+$j]{name},2);
					if ($ID !~ /$AC/)# excellent candidate
					{
						$k = $j;
						last;
					}
					$k=$j if (($k<0)&&($ID !~ /XXXXX/));# good candidate
				}
				$name = $OldSeq[$i+$k]{name} if ($k > 0);
			}
			&AddSequence($n,"+$name",'.','.',$w,$seq);
		}
		else
		{
			push @Sequences, $OldSeq[$i];
		}
	}
	close(CONS);
	system("rm -f $tmp");
	unless ($Consensus)
	{
		$Consensus = $Sequences[0]{alignment};
		$Consensus =~ tr/()[]//d;
		$_ = &putCO($Consensus);
		chomp;
		push @Bottom, $_;
	}
}

sub SubEukaryota
{
	my ($l) = @_;
	if ($l>0)
	{
		if ($l < 1800) { $l = 'R'; }
		elsif ($l < 8000) {$l = 'Z'; }
		elsif ($l < 8600) { $l = 'F'; }
		else { $l = 'L'; }
	}
	return $l;
}

sub lookKingdom
{
	 my ($key) = @_;
	 local ($_);
	 return 0 if ($max1 < 1);
	 my($min, $max, $mid) = (0,$max1 );
	 $key =~ tr/a-z/A-Z/;
	 while ($max - $min > 1)
	 {
		  $mid = int(($max + $min) / 2);
		  if (defined($_= $Key[$mid]) && $_ lt $key)
		  {
				$min = $mid;
		  }
		  else
		  {
				$max = $mid;
		  }
	 }
	 $min=$max if ($Key[$min] lt $key);
	 return "" if $Key[$min] ne $key;
#	 return SubEukaryota($SubVal[$min]) if ($Val[$min] eq 'E');
	 return $Val[$min];
}

sub CommonKingdom
{
	my ($k, $l)=@_;
	return 0; # no SubKingdom
	$_='ERZFL';
	if (/$k/ and /$l/)
	{
		$k = 'E';
	}
	else
	{
		$k = 0;
	}
	return $k;
}

sub AddKingdom
{
	local $/ = "\n";
	local $max1=0;
	&InitExtraInfo;
	splice @Key, 0;
	splice @Val, 0;
	splice @SubVal, 0;
	# read the species table
	open (SPECIES,$SPECIES) or return ;
	while (defined($_= <SPECIES>))
	{
		($Key[$max1],$Val[$max1],$SubVal[$max1])=split;
		$max1++;
	}
	close (SPECIES);

	# edit names so that they match Word1_Word2
	# and find the kingdom of the clusters
	my  ($k, $prev, $Sp, @Fa);
	for $i (0..$#Sequences)
	{
		$Sequences[$i]{name}=correctAC_ID(&getAC_ID($Sequences[$i]{name}));
		if ($Sequences[$i]{nb} < 2)
		{
			$k = lookKingdom(&getSeqSp($Sequences[$i]{name}));
			$prev=$i;
			$Kingdom[$i]=$k;
			while ($k && $prev && ($Sequences[$prev-1]{nb}==1))
			{
				$prev=$Fa[$prev];
				$k = CommonKingdom($k,$Kingdom[$prev+1]) if($k ne $Kingdom[$prev+1]);
				$Kingdom[$prev]=$k;
			}
		}
		else
		{
			$Kingdom[$i]=0;
			$Fa[$i+2*$Sequences[$i+1]{nb}]=$i;
		}
	}
	# edit the names with kingdom info
	for $i (0..$#Sequences)
	{
		$k = $Kingdom[$i];
		if ($k)
		{
			$Sequences[$i]{name} =~ s/(_.*)/$1_$k/;
		}
	}
}

sub isRedundant
{
	my ($new, $list)=@_;
	return 0 if (!$list or (@$list==0));
	my ($old, $old_begin, $old_end);
	my ($new_begin,$new_end) = $new =~ /\((\d+)\-(\d+)\)/;
	foreach $old (@$list)
	{
		($old_begin,$old_end) = $old =~ /\((\d+)\-(\d+)\)/;
		return 1 if (($new_begin >= $old_begin) && ($new_end <= $old_end));
	}
}

sub PDBnr
{
	my %Idx;
	my ($cles, $cle, $dr);
	my $PDB = $DataRelated{PDB};
	for ($i=0; $i < @$PDB; $i++)
	{
		$dr = $$PDB[$i];
		$cles = "$$dr[0]";#$$dr[3]";
		$cle = "$$dr[4]";#$$dr[5]";
#		print STDERR join('#',@$dr)," => $cles#$cle\n";
#		if (!$Idx{$cle} && !$Idx{$cles})
		$Idx{$cles} = [] unless $Idx{$cles};
		$Idx{$cle} = [] unless $Idx{$cle};
		if (!&isRedundant($$dr[3],$Idx{$cles}) && !&isRedundant($$dr[5],$Idx{$cle}))
		{
#				$Idx{$cles}=1;
#				$Idx{$cle}=1;
				push @{$Idx{$cles}}, $$dr[3];
				push @{$Idx{$cle}}, $$dr[5];
		}
		else
		{
			splice @$PDB, $i, 1;
			$i--;
		}
	}
}

sub Common2Complete
{
	($ret_val) = &MakeTree();
	if ($TreeMethod >= 0)
	{
		my @MethodText = ("Bionj","MissDist+Bionj","TriangleTree","Error");
		$_ = &putParsableComment($TREE_METHOD_KEY,$MethodText[$TreeMethod]);
		chomp;
		push @Title, $_;
	}
		  if ($ret_val < 0)
		  {
				return ($ret_val);
		  }
	return 0 unless ($Arbre or ($NbSeq<2));
	&Consens;
	return 0 unless (2*$NbSeq -1 == @Sequences);
	&AddKingdom;
	&PDBnr if $DataRelated{PDB};
	return 1;
}

sub CodeAsCons
{
	my ($Seq, $Cons) = @_;
	my ($Ouvre, $MotCons, $Ferme, $l, $Ident, $MotIdent, $MotDiff, $Motseq);
	my $S = "";
	while ($Cons ne "")
	{
		($MotCons, $Cons) = $Cons =~ /^([^\(\[]*)(.*)/;
		if ($l = length($MotCons))
		{
			($MotSeq, $Seq)= $Seq =~ /(.{$l})(.*)/;
#			print STDERR "$l $MotCons $MotSeq\n";
			$S .= $MotSeq;
		}
#		print STDERR "$S\n";
		last if ($Cons eq "");
		($Ouvre, $MotCons, $Ferme, $Cons) = $Cons=~ /([\(\[])([^\)\]]+)([\)\]])(.*)/;
		if ($l = length($MotCons))
		{
			($MotSeq) = $Seq =~ /(.{$l})/;
			$Ident = ($MotCons ^ $MotSeq);
			while ($Ident ne "")
			{
#$Ident can includes \n because "\n"="L"^"F" for example so / /s is necessary
				($MotIdent, $MotDiff, $Ident) = $Ident =~ /^(\0*)([^\0]*)(.*)$/s;
#				print STDERR join(' ',length($MotIdent),length($MotDiff),length($Ident)),"\n";
				if ($l = length($MotIdent))
				{
					($MotSeq, $Seq) = $Seq =~ /(.{$l})(.*)/;
					$S .= $Ouvre.$MotSeq.$Ferme;
				}
				if ($l = length($MotDiff))
				{
					($MotSeq, $Seq) = $Seq =~ /(.{$l})(.*)/;
#					print STDERR "$l $MotDiff $MotSeq\n";
					$S .= $MotSeq;
				}
			}
		}
	}
	$S;
}

sub getFather
{
	my ($root) = @_;
	my ($n, $m, $x);
	return 0 if ($root < 2) ;
	$n=$Sequences[$root]{nb};
	$x=$root-1;
	if ($x+2*$n < @Sequences)
	{
		$m=$Sequences[$x]{nb};
#		print STDERR "pere=$x ($m) fils2=($x+2*$n) (",$Sequences[$x+2*$n]{nb},") fils1=$root ($n)\n";
		return $x if ($m == $n + $Sequences[$x+2*$n]{nb});
	}
	for ($x=0;$x<$root;)
	{
		$m=$Sequences[$x]{nb};
		if ($x+2*$m-1 < $root)
		{
			$x += 2*$m-1;
		}
		else
		{
#			print STDERR "pere=$x ($m) fils1=($x+1) (",$Sequences[$x+1]{nb},") fils2=$root ($n)\n";
			return $x if ($m == $Sequences[$x+1]{nb} + $n);
			$x++;
		}
	}
	return $root;# error !
}

sub Extract
{
	local ($search) = @_;
	my @list;
	my $Seq;
	foreach $Seq (@Sequences)
	{
		next if ($Seq->{nb} > 1);
		push @list, $Seq if ($Seq->{name}=~/$search/);
	}
	return @list;
}

################
#
# global parameters to define subtrees
#
#	$root			0              root node identificator
#	$maxleaves	12					maximal number of leaves
#	$mindist		20					minimal distance between leaves
#	$search		""					pattern looked for in the clustered names
#	$wanted		align				align|tree choice of the representation
#	$treetype	'p'             'h'|'r'|'p' horiz star, radial star, phenogram
#
# C programs
#  $SUBTREE $id $maxleaves $mindist <$Arbre : compute a subtree for domain $id
#	$DRAWTREE $gif_file $treetype <$tree_file |sort >cf_file:
#		draw the tree described in tree_file (type: $treetype), put the image into
#		gif_file and the node map into cf_file
#
################

sub ValidateParameters
{
	$root = int($root);
	$maxleaves= $NbSeq+0 if (($maxleaves eq "all") or ($maxleaves > $NbSeq));
	$maxleaves = 1 if ($maxleaves <= 0);
	$maxleaves = int($maxleaves);
	$mindist = int($mindist);
	if ($wanted eq 'rtree')
	{
		$treetype = 'r';
	}
	elsif ($wanted eq 'ptree')
	{
		$treetype = 'p';
	}
	elsif ($wanted eq 'htree')
	{
		$treetype = 'h';
	}
	else
	{
		$treetype ='p';
		$wanted = 'align' unless ($wanted =~ /tree/);
	}
}

sub Subtree
{
	my ($r, $l, $d, $file)=@_;
	open (TREE,"| $SUBTREE $r $l $d >$file") or return 0;
	print TREE $Arbre;
	close (TREE);
	return (ExistFile($file));
}

sub Rootmap
{
	local $/ = "\n";
	my ($cf_file, $x, $y) = @_;
	my @coord;
	my ($d,$dx,$dy);
	$root=0;
	open (MAP, $cf_file) or return 0;
	while (defined($_= <MAP>))
	{
		@coord = split;
		last if (($dx=abs($coord[0]-$x))<4);
	}
	$d=4;
	while ($dx < $d)
	{
		if (($dy=abs($coord[1] - $y))<$d)
		{
			$d = ($dx>$dy) ? $dx : $dy;
			$root = $coord[2];
			last if ($d == 0);
		}
		last unless (defined($_= <MAP>));
		@coord = split;
		$dx = abs($coord[0]-$x);
	}
	close(MAP);
	return $root;
}


sub GetName
{
	my ($i) = @_;
	my ($name, @G);
	my $last = $i + 2*$Sequences[$i]{nb}-1;
	$_ = $Sequences[$i]{name};
	s/^\+//;
	@G = split('_',$_,3);
	s/_.$//;
	$name = $_;
	if ($search ne "")
	{
		for ($i++;$i<$last;$i++)
		{
			$_ = $Sequences[$i]{name};
			if (/$search/)
			{
				s/^\+//;
				s/_.$//;
				$name = $_;
				last;
			}
		}
	}
	return ($name,$G[2]) if $G[2];
	$name;
}

sub Drawtree
{
	local $/ = undef;
	my ($gif_file,$cf_file,$tree_file) = @_;
	my ($i, $name);
	open (INTREE, $tree_file) or return 0;
#	open (OUTTREE,">$gif_file");
	open (OUTTREE,"|$DRAWTREE $gif_file $treetype|sort >$cf_file")
		or return 0;
# test
#	open (TEST, ">subtree.$Id_fam");
	while (defined($_= <INTREE>))
	{
		s/(\d+)/&getSeqID(join('_',&GetName($1)))/ge;
		print OUTTREE $_;
#		print TEST $_;
	}
	close(INTREE);
#	close(TEST);
	return close(OUTTREE);
}

sub ListSeq
{
	my ($first, $last, $Seq, @list);
	$first = $root;
	$last = $first + 2*$Sequences[$root]{nb}-2;
	foreach $Seq (@Sequences[$first .. $last])
	{
		next if ($Seq->{nb} > 1);
		push @list, &getSeqAC($Seq->{name});
	}
	return @list;
}

sub FetchFile
{
	my ($id_dom) = @_;
	local $_ = $FETCHFAM;
	s/ID_FAM/$id_dom/;
	#print STDERR "## $_ | ## \n";
	return "$_ |";
}

sub FetchDomain
{
	my ($id_dom) = @_;
	return 0 unless (&ReadDomain (&FetchFile($id_dom)) and &VerifyDomain);
	unless ($Id_fam)
	{
		$Id_fam = $id_dom ;
		$_ = &putID($Id_fam);
		chomp;
		unshift @Title, $_;
	}
#	print STDERR "domain $Id_fam : $NbSeq sequences de $Length\n";
}

sub CreateProdomCompleteDomain
{
	my ($id_dom1, $complete) = @_;
	my ($ret_val) = (0);
#	print STDERR "create ${DOMDIR}$id_dom1\n";
	if(&ExistFile(${DOMDIR}.$id_dom1))
	{
#		print STDERR "${DOMDIR}$id_dom1 existe\n";
		system ("touch ${DOMDIR}$id_dom1");
		return 1;
	}
	return -1 unless (&ValidDom($id_dom1));
	return -2 unless (&FetchDomain($id_dom1));
	return 1 unless ($complete);
	$ret_val=&Common2Complete;
	if ($ret_val>0)
	{
#		print STDERR "saved to ${DOMDIR}$id_dom1\n";
		&PrintDomain(">${DOMDIR}$id_dom1");
	}
	elsif ( $ret_val < 0 )
	{
		return ($ret_val);
	}
		  $ret_val = &ExistFile("${DOMDIR}$id_dom1");
	return ($ret_val);
}

sub ReadProdomDomain
{
	my ($id_dom, $complete) = @_;
	my ($ret_val) = (0);
	&InitRelease;
#	return 0 unless &CreateProdomCompleteDomain($id_dom);
	($ret_val) = &CreateProdomCompleteDomain($id_dom, $complete);
	if ( $ret_val <= 0 )
	{
		return($ret_val);
	}

	if ( (!defined($Id_fam)) or ($Id_fam!=$id_dom) )
	{
		&ReadDomain("${DOMDIR}$id_dom");
#		print STDERR "domain $Id_fam : $NbSeq sequences de $Length\n";
	}
	return (($Id_fam == $id_dom) and &VerifyDomain);
}

sub ReadProdomCompleteDomain
{
	my ($id_dom) = @_;
	return &ReadProdomDomain($id_dom, 1);
}


##############################
#### 3: a common part
##############################
#### pseudo code
####		Initialisation
####		unless ReadFamily
####			print Sorry
####			die
####
####		printBigTitle
####		if ($NbSeq > 2)
####			ValidateParameters
####			SelectRoot
####			UpdateNames
####			if ($root>0)
####				$Subtitle= "Subfamily: ..."
####
####			if ($wanted eq "tree")
####				printTree
####			else
####				printAlignment
####		else
####			printSmallAlignment
#######################

########## HTML
if ($CSS)
{
	$STYLE = "<style type='text/css'>\n".
		"body {background: silver}\n".
		"table.title {background: white}\n".
		"th.title {font-size: larger}\n".
		"table.comment {background: white}\n".
		"td.AC {font-size: larger}\n".
		"a.B {color: blue}\n".
		"a.P {color: blue}\na.A {color: red }\na.E {color: yellow}\n".
		"a.V {color: lime}\na.L {color: olive}\na.Z {color: aqua}\n".
		"a.F {color: maroon}\na.R {color: navy}\na.black {color: black}\n".
		"tt.low {color: blue}\ntt.high {color: red}\ntt.no {color: black}\n".
		"img.icon {vertical-align: middle; height: 30}\n".
		"ul.redball {list-style: url(\"$HTTPIMAGE/redball.gif\") disc}\n".
		"td.number {text-align: right}\n".
		"td.rule {text-align: right; white-space: nowrap}\n".
		"td.sequence {white-space: nowrap}\n".
		"th.sequence {white-space: nowrap}\n".
		"</style>";
	$BODY = "";
	$BigTitleFormat="<center><table class=title border=1><tr><th class=title>TITLE<th>INFO</tr></table></center>";
	$CommentTableBegin="<table class=comment border=2>";
#	$AClineFormat="<tr><td>Accession number<td class=AC>PD000000</tr>\n";
	$AClineFormat="<tr><td>Accession number<td class=AC>PD000000/MOTIFGIF</tr>\n";
	$tt{low}="<tt class='low'>";
	$tt{high}="<tt class='high'>";
	$tt{'no'}="</tt>";
	$spform="class";
	$spref=Ahref("GETREF","COL","NAME")."\n";
	$midpos="class=icon";
	$rulemask1="<td class=rule><tt>";
	$rulemask11="<td class=sequence><tt>";
	$seqheadermask="<table cellpadding=0><thead><tr><th><th><th>Sequence ID"
		."<th>start<th>end<th>weight<th class=sequence>RULE</tr><tbody>\n";
	$seqlinemask="<tr><td class=number>NB<td>PLUS<td>NAME<td class=number>START"
		."<td class=number>END<td class=number>WEIGHT"
		."<td class=sequence><tt>SEQ</tt></tr>";
}
else
{
	$STYLE="";
	$BODY = "bgcolor=silver";
	$BigTitleFormat="<center><table  border=1 bgcolor=white><tr><th><font size='+2'>TITLE</font><th>INFO</tr></table></center>";
	$CommentTableBegin="<TABLE BORDER=2 ALIGN=CENTER BGCOLOR=white>\n";
#	$AClineFormat="<TR><TD>Accession number<TD><Font SIZE='+1'>PD000000</font></TR>\n";
	$AClineFormat="<TR><TD>Accession number<TD><Font SIZE='+1'>PD000000/</font>MOTIFGIF</TR>\n";
	$Color{E}="yellow";
	$Color{A}="red";
	$Color{B}="blue";
	$Color{P}="blue";
	$Color{V}="lime";
	$Color{L}="olive";
	$Color{Z}="aqua";
	$Color{F}="maroon";
	$Color{R}="navy";
	$tt{low}="<font color=blue>";
	$tt{high}="<font color=red>";
	$tt{'no'}="</font>";
	$spform="color";
	$spref=Ahref("GETREF","","<font COL>NAME</font>")."\n";
	$midpos="align=middle height=30";
	$rulemask1="<td align='right' nowrap><tt>";
	$rulemask11="<td nowrap><tt>";
	$seqheadermask="<table cellpadding=0><thead><tr><th><th><th>Sequence ID"
		."<th>start<th>end<th>weight<th nowrap>RULE</tr><tbody>\n";
	$seqlinemask="<tr><td align='right'>NB<td>PLUS<td>NAME<td align='right'>START"
		."<td align='right'>END<td align='right'>WEIGHT"
		."<td nowrap><tt>SEQ</tt></tr>";
	$Subtitle2Mask="<table border=1 bgcolor=white><tr>"
		."<td>Other descriptions of the SUBfamily<br>corresponding to this WANTED"
		."<td>WANTEDREF PERSO"
		."</table>";
}
$noborder="border=noborder";
$rulemask0="<table border='0' cellspacing='0' cellpadding='0'><tr>\n";
$rulemask2="</tt></td>\n";
$rulemask3= "</tr><tr>\n";
$rulemask4= "</tr></table>\n";

sub getColor
{
  my ($Col)=@_;
  if ($Col)
  {
		$Col = $Color{$Col} if (!$CSS);
		return "$spform=$Col";
  }
  else
  {
		return "$spform=black";
  }
}

$Separator ="<hr>\n";

sub MultiLines
{
	join('<br>',@_);
}

sub ParHeader
{
	"<P><H3>".Image("$HTTPIMAGE/redball.gif","","-")
		. join('<br>',@_)."</H3>\n";
}

sub SimpleText
{
	"<P>".join('<br>',@_)."</P>";
}

sub HtmlCenter
{
	"<center>".join("\n",@_)."</center>";
}

sub CenterText
{
	HtmlCenter("<h1>".join('<br>',@_)."</h1>");
}

sub ColorText
{
	my ($text,$col)=@_;
	"<font $col>$text</font>";
}

$TableBegin = "<table>";
$Table1Begin = "<table border=1>";
$TableEnd = "</table>";
sub TableHeader
{
	join ('<th>','<tr>',@_)."\n";
}
sub TableLine
{
	join ('<td>','<tr>',@_);
}

sub TableRow
{
	my ($h, @d) = @_;
	join('<td>',"<tr><th>$h",@d);
}
sub ListBegin
{
	($ListHeader) = @_;
	$TableBegin;
}
$ListEnd = $TableEnd;
sub ListElem
{
	TableRow ($ListHeader,@_);
}

sub Aname
{
	my ($ref, $text) = @_;
	return "<A name = \"$ref\">$text</A>\n";
}

sub Ahref
{
	my ($ref, $comp, $text) = @_;
	return "<A HREF = \"$ref\" $comp>$text</A>\n";
}

sub Image
{
	my ($ref, $comp, $alt) = @_;
	return "<img src=\"$ref\" $comp alt=\"$alt\">";
}

sub FormBegin
{
  join("", "<form method=post action=\"", @_,"\">\n");
}
$FormEnd = "</form>\n";
sub SubmitB
{
	my ($name) = @_;
	"<input type=submit".(($name)? " value='$name'": "") .">";
}
sub TextParam
{
	my ($name, $value, $size)=@_;
	"<input id='$name' type=text name='$name' value='$value' size=$size maxlength=$size>";
}

sub Select
{
	my ($name, $size, $selected, @options) = @_;
	my ($tmp, $i);
	$tmp = "<select name=$name size=$size>";
	for ($i=0;$i<@options;$i++)
	{
		if ($options[$i] ne "")
		{
			$tmp.="<option value=$options[$i]";
			$tmp.=" selected" if ($i == 2*$selected);
			$tmp.=">$options[++$i]</option>";
		}
		else
		{
			$i++;
		}
	}
	$tmp.="</select>";
}

sub HiddenParam
{
	my ($name, $value) = @_;
	return "<input name='$name' type='hidden' value='$value'>\n";
}

sub ImageParam
{
	my ($name, $src, $alt)=@_;
	"<input name='$name' type=image src='$src' alt='$alt'>\n";
}

sub printTableHeader
{
	my ($rule)= @_;
	local $_ = $seqheadermask;
	s/RULE/$rule/;
	print STDOUT $_;
}

sub printTableLine
{
	my ($Nb,$plus,$name,$start,$end,$weight,$seq)= @_;
	local $_ = $seqlinemask;
	s/NB/$Nb/;
	s/PLUS/$plus/;
	s/NAME/$name/;
	s/START/$start/;
	s/END/$end/;
	s/WEIGHT/$weight/;
	s/SEQ/$seq/;
	print STDOUT $_;
}

sub printHeaders
{
	my ($rule, $i);
	local $_;
	$rule=$rulemask0;
	$i = 10;
	while ($i <= $Length)
	{
		$rule .= "$rulemask1$i$rulemask2";
		$i += 10;
	}
	$rule .= $rulemask3;
	$_ = "-"x$Length;
	s/\-{10}/---------|$rulemask2$rulemask11/g;
	$rule .= "$rulemask11$_$rulemask2";
	$rule .= $rulemask4;
	printTableHeader ($rule);
}

sub colorSequence
{
	my ($Seq, $Cons) = @_;
	$Seq=~tr/()[]//d;
	my $S = &CodeAsCons($Seq,$Cons);
	$S=~s/\)\(//g;
	$S=~s/\]\[//g;
	$S=~s/\(/$tt{'low'}/g;
	$S=~s/\[/$tt{'high'}/g;
	$S=~s/\)/$tt{'no'}/g;
	$S=~s/\]/$tt{'no'}/g;
	$S;
}
###################### end of HTML functions

sub FSetDefaultParam
{
	$id_fam = "";
	$maxleaves=12;
	$mindist=20;
	$refx=-1;
	$root=0;
	$search="";
	$wanted="align";
}

sub FUpdateParam
{
	local ($key,$_) = @_;
	if ( $key =~ /id_fam/ )
	{
			$id_fam = "$_";
	}
	elsif ( $key =~ /maxleaves/ )
	{
			$maxleaves = "$_";
	}
	elsif ( $key =~ /mindist/ )
	{
			$mindist = "$_";
	}
	elsif ( $key =~ /oldname/ )
	{
			$oldname = "$_";
	}
	elsif ( $key =~ /ref\.x/ )
	{
			$refx = "$_";
	}
	elsif ( $key =~ /ref\.y/ )
	{
			$refy = "$_";
	}
	elsif ( $key =~ /wanted/ )
	{
			$wanted = "$_";
	}
	elsif ( $key =~ /root/ )
	{
			$root = "$_";
	}
	elsif ( $key =~ /search/ )
	{
			$search = "$_";
			$search =~ tr/a-z/A-Z/;
	}
}

sub InitProg
{
	chdir("$PROG");
	$SUBTREE=$PROG."subtree";
	$DRAWTREE=$PROG."drawtree";
}

sub Script
{
	my ($i, $out, $all) = @_;
	my ($ret);
	$ret = "$SUBTREEPL?".&addSpecialParam()."id_fam=$Ac_fam&amp;mindist=";
	if ($all)
	{
		$ret .= "0&amp;maxleaves=all";	
	}
	else
	{
		$ret .= "$mindist&amp;maxleaves=$maxleaves";
		$ret .= "&amp;search=$search" if ($search);
	}
	return $ret."&amp;wanted=$out&amp;root=$i";
}

sub printSequence
{
	my ($i) = @_;
	my $Seq = $Sequences[$i];
	my ($tname, $plus, $col, $AC);
	local $_;
	($tname,$col)=&GetName($i);
	($AC,$tname) = &getAC_ID($tname);
#   print STDERR "$AC $tname $col\n";
	$_ = $spref;
	s/COL/&getColor($col)/e;
	s/NAME/$tname/;
	s/GETREF/&getRef($AC)/e;
#	$tname = Ahref(&getRef($AC),"",&ColorText($tname,$col));
	$tname = $_;
	$plus ="";
	if ($Seq->{nb} > 1)
	{
		$plus =  &Ahref(&Script($i,'alignment'),"", &Image("$HTTPIMAGE/plus.gif",
			"width=9 height=11 $noborder","subfamily display"));
		$plus .= "\n";
	}
	printTableLine ($Seq->{nb},$plus,$tname,$Seq->{begin},$Seq->{end},
		$Seq->{weight},colorSequence($Seq->{alignment},$Cons));
}

sub printForm
{
	my ($istree) = @_;
	my %selection =(t=>2,ht=>0,rt=>1,pt=>2);
	printSpecialForm();
	print STDOUT HiddenParam("id_fam",$Ac_fam);
	print STDOUT $Table1Begin;
	print STDOUT &TableHeader("Minimal distance between sequences (in PAM)",
		&TextParam('mindist',$mindist,4));
	print STDOUT &TableHeader("Maximal number of clusters",
		&TextParam('maxleaves',$maxleaves,4));
	print STDOUT &TableHeader("If possible, clusterIDs should contain the following string (e.g.: human)",
		&TextParam('search',$search,13));
	if ($istree)
	{
		 print STDOUT &TableHeader("Type of tree or labels",
			&Select('wanted',1,$selection{$treetype.'t'},'htree','Radial tree (horiz. labels)',
				'rtree','Radial tree and labels', 'ptree','Phenogram'));
	}
	print STDOUT "$TableEnd\n";
}

sub printTree
{
	my ($src) = $DOMDIR.$oldname.".gif";
	my ($hsrc) = $HTTPDATA.$oldname.".gif";
	my ($father)= &getFather($root);
	if (&ExistFile($src))
	{
		system("touch ${DOMDIR}${oldname}.*");
	}
	elsif (!&Drawtree ($src,"${DOMDIR}${oldname}.cf",$DOMDIR.$treename))
	{
		print STDOUT &SimpleText("Sorry, can't draw the tree");
		return;
	}
	print STDOUT $Subtitle;
	print STDOUT $Subtitle2;
	print STDOUT $Separator;
	print STDOUT &FormBegin($SUBTREEPL);
	print STDOUT &ImageParam('ref',$hsrc,'Click on the root of a subfamily to display subtree');
	print STDOUT $Table1Begin;
	print STDOUT &TableHeader(
		'tree root',
		'Eukaryotic subfamily root',
		'Bacterial subfamily root',
		'Archaeal subfamily root',
		'Viral subfamily root',
		'Root of a subfamily that goes across domains of life'
	);
	print STDOUT &TableHeader(
		Image("$HTTPIMAGE/triangle.gif","","triangle"),
		Image("$HTTPIMAGE/yellowsquare.gif","","yellow square"),
		Image("$HTTPIMAGE/bluesquare.gif","","blue square"),
		Image("$HTTPIMAGE/redsquare.gif","","red square"),
		Image("$HTTPIMAGE/greensquare.gif","","green square"),
		Image("$HTTPIMAGE/blacksquare.gif","","black square"),
	);
	print STDOUT "$TableEnd\n";
	print STDOUT &SimpleText("The number of sequences in each subfamily is indicated after the ID");
	print STDOUT $Separator;
	print STDOUT &HiddenParam("oldname",$oldname);
	&printForm(1);
	print STDOUT &SimpleText(
		"To display a new tree with these parameters, select a root \n".
			&Select ('root',1,0,$root,'current',($father)?($father,'up one node'):("",""),
				($root)?('0','global'):("",""))." and ".&SubmitB,
		"or click on the root of a subfamily directly in the tree.\n");
	print STDOUT $FormEnd;
}

sub CreateCfFile
{
	local ($treetype,$search,$treename);# uses old values
	if (ExistFile("${DOMDIR}${oldname}.cf"))
	{
		system("touch ${DOMDIR}${oldname}.*");
	}
	else {
		my @Param = split('_',$oldname);
		$treename = join ('_',@Param[0..3]);
		$treetype = $Param[5];
		$search = $Param[4];
		if (ExistFile($DOMDIR.$treename))
		{
			system("touch ${DOMDIR}$treename");
		}
		else
		{
			return 0 unless Subtree(@Param[1..3],$DOMDIR.$treename);
		}
		return 0 unless &Drawtree("${DOMDIR}${oldname}.gif","${DOMDIR}${oldname}.cf",$DOMDIR.$treename);
	}
	return 1;
}

sub UpdateOldname
{
# update oldname & treename with new parameter values
	my $OK = 1;
	$treename = join ('_',$id_fam,$root,$maxleaves,$mindist);
	if (&ExistFile($DOMDIR.$treename))
	{
		system ("touch ${DOMDIR}$treename");
	}
	else
	{
		$OK = &Subtree($root,$maxleaves,$mindist,$DOMDIR.$treename);
	}
	$oldname = join ('_',$treename,$search);
	$oldname .= "_$treetype" if ($treetype ne '');
	return $OK;
}

sub SelectRoot
{
	$root=0;
	$root = &Rootmap ("${DOMDIR}${oldname}.cf",$refx,$refy)
		if &CreateCfFile;
}

sub printAlignment
{
	local $/ = undef;
	my $name;
	print STDOUT $Subtitle, $Subtitle2;

	printHeaders();
	unless (open (TREE, $DOMDIR.$treename))
	{
		print STDOUT &SimpleText("Sorry, can't read the alignment");
		return;
	}
	while (defined($_=<TREE>))
	{
		foreach $name (&getSubfamily($_))
		{
			&printSequence($name);
		}
	}
	close(TREE);
	print STDOUT &TableRow("");
	printTableLine ($NbCons,"","Consensus","","",$WCons,&colorSequence($Cons,$Cons));
	printSites;
	print STDOUT "$TableEnd $Separator";
	print STDOUT &FormBegin($SUBTREEPL);
	print STDOUT HiddenParam("wanted","alignment");
	printForm(0);
	print STDOUT &MultiLines("To display a new alignment with these parameters, ");
	if ($root > 0)
	{
		print STDOUT &MultiLines("select a root \n".&Select ('root',1,0,$root,'current','0','global')." and ");
	}
	else
	{	
		print STDOUT HiddenParam("root",$root);
	}
	print STDOUT &MultiLines(&SubmitB(),
		"or click ".&Ahref(&Script($root,"alignment",1),"","here")." to  display all sequences in the ${sub}family"),
		$FormEnd;
}

sub printSmallAlignment
{
	$root=0;
	$Cons=$Sequences[0]{alignment};
	$NbCons=$Sequences[0]{nb};
	$WCons =$Sequences[0]{weight};
	printHeaders();
	if ($NbSeq > 1)
	{
		foreach $i (1..2)
		{
			printSequence ($i);
		}
		print STDOUT &TableRow("");
		printTableLine ($NbCons,"","Consensus","","",$WCons,&colorSequence($Cons,$Cons));
	}
	else
	{
		printSequence (0);
	}
	printSites;
	print STDOUT $TableEnd;
#	printBottom();
}

sub WantedRef
{
	my ($script, $target, $gif)=@_;
	&Ahref(&Script($root,$script),"target='$Project${target}_$Ac_fam'",Image("$HTTPIMAGE/$gif.gif",$midpos,$script));
}


###################################################################
##
## main subroutine
##
###################################################################
sub FDomDisplay
{
	my ($name);
	my $init = ($wanted =~ /descript/);
	local ($sub)='';
	&InitFlo ();
	&InitProg();
	unless (&ReadProdomDomain($id_fam, ($init==0))>0)
	{
		print STDOUT &SimpleText("$Separator <H3>$Sorry </H3>"), "</BODY></HTML>";
		die;
	}
	if ($init)
	{
		printBigTitle() ;
		return;
	}
	if ($NbSeq > 2)
	{
		&ValidateParameters;
		&SelectRoot if ($refx>=0);
		unless (&UpdateOldname)
		{
				print STDOUT &SimpleText("$Separator <H3>$Sorry </H3>"), "</BODY></HTML>";
				die;
		}

		$Cons=$Sequences[$root]{alignment};
		$NbCons=$Sequences[$root]{nb};
		$WCons =$Sequences[$root]{weight};
		$Subtitle2 = $Subtitle2Mask;
		$name = ($root>0)?getSeqID(GetName($root)):"";
		$sub = 'sub' if ($name);
		$Subtitle2 =~ s/SUB/$sub/;
		&setSubtitles($name);

		if ($wanted =~ /tree/)
		{
			$Subtitle2 =~ s/WANTED/tree/;
			$Subtitle2 =~ s/WANTEDREF/&WantedRef("alignment","Alignment","alignement")/e;
			&printTree;
		}
		else
		{
			$Subtitle2 =~ s/WANTED/alignment/;
			$Subtitle2 =~ s/WANTEDREF/&WantedRef("tree","tree","tree")/e;
			&printAlignment;
		}
	}
	else
	{
		&printSmallAlignment;
	}
}

###### 4: example of main

#####################################################
#### DisplayFam.pl can work in a cgi-bin directory or locally
#### syntaxe : cgi-bin : DisplayFam.pl?param1=value1&param2=value2.....
####	       locally : DisplayFam.pl param1=value1 param2=value2....."
####
#### These parameters must be initialised in the main program
#### 	parameter	default_value	meaning
####    id_fam		no default	family identificator used by FETCHFAM
####	root		0		root node identificator
####	maxleaves	12		maximal number of leaves
####	mindist		20		minimal distance between leaves
####	search		""		pattern looked for in the clustered names
####	wanted		align		choice of the representation
####					align : two-framed window with description and alignment
####					descript : family description
####					alignment : multiple alignment
####					?tree : tree as horiz star(?=h), radial star(?=r), phenogram(?=p or nothing)
####	oldname		no default	map file name
####	ref.x		-1		x coordinate of the map click (or -1 for no click)
####	ref.y		no default	y coordinate of the map click
####
#####################################################
require 'cgi-lib2.pl';

sub FInitVar
{

  my ($ret,       # Return value of the ReadParse call.
		$buf,			# Buffer for data read from disk.
		$key, $val  # One parameter read as $key=$val
	  );

  # Set default values for all parameters
  &FSetDefaultParam();

  # Start off by reading and parsing the data.  Save the return value.
  $ret = &ReadParse();

  # A bit of error checking never hurt anyone
  if (!defined $ret) {
	 &CgiDie("Error in reading and parsing of CGI input");
  } elsif (!$ret) {
	 &CgiDie("Missing parameters\n",
				"Please complete the <a href='".&MyFullUrl()."'>form</a>.\n");
  }
#	print &PrintHeader();
#	print &PrintVariables();

	# set parameter values
	while (($key, $val) = each(%in))
	{
		&FUpdateParam ($key, $val);
		if (( $key =~ /alignment/ )||( $key =~ /msf/ ) and ($val ne ""))
		# the first time the alignment is in the form
		{
			if ($val =~ /\n/)
			{
				$Alignment = $val;
				$Alignment =~ s/\r//g;
			}
			else
			{
			$Alignment ="";
			# Read the contents of the local file
			open (UPFILE, $val) or next;
			$Alignment .= $buf while (read (UPFILE, $buf, 8192));
			close (UPFILE);
			}
		}
	}
	# set the html page title
	$title = "DisplayFam from F.Corpet (c) I.N.R.A. 1999";
}

sub FInitDocument
{
	select(STDOUT);
	$|=1;
	print &PrintHeader;#cgi-lib function
	print STDOUT "<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 4.0 Transitional//EN\"\n";
	print STDOUT "\"http://www.w3.org/TR/REC-html40/loose.dtd\">";
	print STDOUT "\n<HTML><HEAD><TITLE>$title</TITLE>$STYLE</HEAD>\n";
}

sub FPrintTail
{
		print STDOUT "</BODY></HTML>\n";
}

### comment the sub line to use this subroutine as main
### if you put your main code in another file, be careful to add
### require 'this_file.pl'

#sub Fmain_example
{
	&FInitVar();
	&FInitDocument();
	if ($id_fam ne "")
	{
		&FDomDisplay();
	}
	else
	{
		print STDOUT "<P><H3>Sorry, your request is not valid.</H3></P>\n";
	}
	&FPrintTail();
  # The following lines are solely to suppress 'only used once' warnings
  $ListEnd=$ListEnd;
  %in=%in;
}

